from .. import Scraper
from ravenclaw import standardize_columns
from pandas import DataFrame
from silverware import read_table


class Country:
	def __init__(self, name, alpha_2, alpha_3, numeric):
		self._dictionary = {
			'name': name,
			'alpha_2': alpha_2,
			'alpha_3': alpha_3,
			'numeric': numeric
		}

	def __getstate__(self):
		return self._dictionary

	def __setstate__(self, state):
		self._dictionary = state

	@property
	def name(self):
		return self._dictionary['name']

	@property
	def alpha_2(self):
		return self._dictionary['alpha_2']

	@property
	def alpha_3(self):
		return self._dictionary['alpha_3']

	@property
	def numeric(self):
		return self._dictionary['numeric']

	def get(self, key):
		return self._dictionary[key]

	def __eq__(self, other):
		return self.alpha_3 == other.alpha_3

	def __lt__(self, other):
		return self.name < other.name

	def __gt__(self, other):
		return self.name > other.name

	def __le__(self, other):
		return not (self > other)

	def __ge__(self, other):
		return not (self < other)

	def __ne__(self, other):
		return not (self == other)


class CountryCodes(Scraper):
	def __init__(self, rate_limit_wait_seconds=0.01, cache=None, num_request_tries=4):
		self._table = None
		self._countries = None
		super().__init__(
			name='country_codes', expire_in='1 year', rate_limit_wait_seconds=rate_limit_wait_seconds,
			cache=cache, num_request_tries=num_request_tries
		)

	def _create_cached_functions(self):
		super()._create_cached_functions()
		if self.cache:
			self.get_table = self.cache.make_cached(
				id='iban_country_codes_get_table',
				function=self._get_table,
				sub_directory='iban_cc_table',
				expire_in='1 year'
			)
		else:
			self.get_table = self._get_table

	def _get_state_attribute_names(self):
		return super()._get_state_attribute_names() + ['_table', '_countries']

	def _get_table(self):
		"""
		:rtype: DataFrame
		"""
		soup = self.get_soup(url='https://www.iban.com/country-codes')
		return standardize_columns(data=read_table(table=soup.find(name='table'))).rename(
			columns={'country': 'name', 'alpha_2_code': 'alpha_2', 'alpha_3_code': 'alpha_3'}
		)

	@property
	def table(self):
		"""
		:rtype: DataFrame
		"""
		if self._table is None:
			self._table = self.get_table()
		return self._table

	@property
	def countries(self):
		"""
		:rtype: dict[str, Country]
		"""
		if self._countries is None:
			data = self.table.copy()
			countries = data.apply(
				func=lambda row: Country(
					name=row['name'], alpha_2=row['alpha_2'], alpha_3=row['alpha_3'], numeric=row['numeric']
				),
				axis=1
			)
			self._countries = {
				getattr(country, attribute): country
				for country in countries
				for attribute in ['name', 'alpha_2', 'alpha_3', 'numeric']
			}
		return self._countries

	def convert(self, code, to='alpha_3'):
		return self.countries[code].get(key=to)



"""
from selenium.webdriver.chrome.webdriver import WebDriver as ChromeDriver
from selenium.webdriver.firefox.webdriver import WebDriver as FirefoxDriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.common.by import By
from selenium.webdriver import Chrome, Firefox
"""
import requests
from requests.adapters import SSLError
from bs4 import BeautifulSoup

import time
import warnings
from chronometry import MeasurementSet


class Navigator:
	def __init__(self, driver=None, user_agent=None, request_method='urllib', timeout=10):
		"""
		:type driver: str or ChromeDriver or FirefoxDriver or NoneType
		:param str user_agent: the default user agent, one of random, ie, ff, chrome, etc.
		"""

		'''
		if isinstance(driver, str):
			if driver.lower() == 'chrome':
				driver = Chrome()
			elif driver.lower() == 'firefox':
				driver = Firefox()
			else:
				raise ValueError(f'Unknown driver: "{driver}"')
		else:
			if driver is None and request_method != 'urllib':
				raise ValueError(f'Driverless Navigator only works with urllib method, not with "{request_method}"!')
		'''
		self._driver = driver

		try:
			self._user_agent = user_agent.random
		except:
			self._user_agent = 'Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.103 Mobile Safari/537.36'

		self._default_request_method = request_method
		self._timeout = timeout
		self._url = None
		self._page_source = None
		self._parsed_html = None
		self._measurement_set = MeasurementSet()

		#self.get = self._measurement_set.measure(function=self._get, name='get')
		#self.parse_html = self._measurement_set.measure(function=self._parse_html, name='parse_html')
		self.get = self._get
		self.parse_html = self._parse_html

	def __del__(self):
		self.driver.quit()

	@property
	def driver(self):
		"""
		:rtype: ChromeDriver or FirefoxDriver
		"""
		return self._driver

	def _get_by_driver(self, url, element_id=None, timeout_exception='error'):
		self.driver.get(url=url)
		if element_id is None:
			time.sleep(self._timeout)
		else:
			try:
				element_present = expected_conditions.presence_of_all_elements_located((By.ID, element_id))
				WebDriverWait(driver=self.driver, timeout=self._timeout).until(element_present)

			except TimeoutException:
				self._page_source = None
				if timeout_exception[0].lower == 'e':
					raise TimeoutException(f'Timed out waiting for page:"{url}" to load!')
				elif timeout_exception[0].lower == 'w':
					warnings.warn(message=f'Timed out waiting for page:"{url}" to load!')
		return self.driver.page_source

	def _get_by_urllib(self, url, headers=None, json=False):
		headers = headers or {
			'user-agent': self._user_agent
		}
		request = requests.get(url=url, headers={'User-Agent': self._user_agent})
		'''
		request = urllib.request.Request(url)
		
		#for key, value in headers.items():
			request.add_header(key=key, val=value)

		if json:
			response = urllib.request.urlopen(request, timeout=self._timeout)
			return load_json(response)
		else:
			with urllib.request.urlopen(request, timeout=self._timeout) as response:
				html = response #.read().decode(encoding)
			return html
		'''
		if format == 'json':
			try:
				r = requests.get(url, params=parameters, headers=headers)
				result = r.json()
			except SSLError as e:
				warnings.warn(str(e))
				result = None

	def _get(
			self, url, request_method=None, element_id=None, get_json_back=False,
			timeout_exception='error', parser='lxml'
	):
		"""
		:type url: str
		:param str or NoneType request_method: one of 'urllib' or 'selenium' or None, None will choose the default
		:param str or NoneType user_agent: one of None (to choose default), random, ie, ff, etc.
		:type element_id: str or NoneType
		:type encoding: str
		:rtype: BeautifulSoup or str
		"""
		self._url = url
		request_method = request_method or self._default_request_method

		if request_method.lower() == 'urllib':
			html = self._get_by_urllib(url=url, json=get_json_back)

		elif request_method.lower() == 'selenium':
			html = self._get_by_driver(url=url, element_id=element_id, timeout_exception=timeout_exception)

		else:
			raise ValueError(f'Unknown method: "{method}"!')

		self._page_source = html

		if parser:
			return self.parse_html(parser=parser, html=html)

		else:
			return self._page_source

	def _parse_html(self, html, parser='lxml'):
		"""
		:type parser: str
		:type html: str
		:rtype: BeautifulSoup
		"""
		self._parsed_html = BeautifulSoup(html, parser)
		return self._parsed_html

	@property
	def loading_time(self):
		return self._get_timer.get_mean_elapsed(unit='sec')

	@property
	def parsing_time(self):
		return self._parse_timer.get_mean_elapsed(unit='sec')



import json
from silverware import separate_row_header, parse_link


import re

WORD = re.compile(r'\w+')


def tokenize(text):
	"""
	this function tokenizes text at a very high speed
	:param str text: text to be tokenized
	:rtype: list[str]
	"""
	words = WORD.findall(text)
	return words


class InfoBox:
	def __init__(self, html, extract=False):
		#strainer = SoupStrainer('table', {'class': re.compile('infobox.+vcard')})
		if html:
			for br in html.find_all('br'):
				br.replace_with('\n')
			self._dictionary = self._parse_table(html)
			if extract:
				html.extract()
		else:
			self._dictionary = {}

	def __str__(self):
		return '\n'.join([f'{k}: {v}' for k, v in self._dictionary.items()])

	def __repr__(self):
		return 'InfoBox:\n'+str(self)

	def __getstate__(self):
		return self._dictionary

	def __setstate__(self, state):
		self._dictionary = state

	def __getitem__(self, item):
		return self._dictionary[item]

	def __contains__(self, item):
		return item in self._dictionary

	def copy(self):
		duplicate = self.__class__(html=None)
		duplicate._dictionary = self._dictionary.copy()
		return duplicate

	@staticmethod
	def _parse_table(table):
		result = {}
		title_number = 1
		unknown_header_number = 1

		for row in table.find_all('tr'):
			header, rest = separate_row_header(row)

			texts = ['_'.join(tokenize(text)) for text in rest.text.split('\n') if text != '']
			links = [parse_link(link) for link in rest.find_all('a')]

			if len(texts) > 0 or len(links) > 0:
				if header:
					header_text = '_'.join(tokenize(header.text))
				else:
					header_text = f'unknown_row_{unknown_header_number}'
					unknown_header_number += 1

				if header_text in result:
					result[header_text]['texts'] += texts
					result[header_text]['links'] += links
				else:
					result[header_text] = {'texts': texts, 'links': links}
			elif header:
				result[f'title_{title_number}'] = '_'.join(tokenize(header.text))
				title_number += 1

		return result

	def items(self):
		return self._dictionary.items()

	def keys(self):
		return self._dictionary.keys()

	def values(self):
		return self._dictionary.values()



from pensieve import Pensieve
from chronometry.progress import ProgressBar
from silverware import Link


def get_special_data(wikipedia, name, echo=1):
	"""
	:type wikipedia: .Wikipedia.Wikipedia
	:type name: str
	:rtype: list[Pensieve]
	"""
	if name == 'country_pages':
		page = wikipedia.get_page(url='https://en.wikipedia.org/wiki/List_of_national_capitals')
		countries = list(page['tables'][0]['country'])

		for one_or_more_links in countries:
			if isinstance(one_or_more_links, list):
				for link in one_or_more_links:
					if isinstance(link, Link):
						yield wikipedia.get_page(url=link.url)
			elif isinstance(one_or_more_links, Link):
				yield wikipedia.get_page(url=one_or_more_links.url)

	if name == 'country_capital_pages':
		page = wikipedia.get_page(url='https://en.wikipedia.org/wiki/List_of_national_capitals')
		capitals = list(page['tables'][0]['city'])

		for one_or_more_links in capitals:
			if isinstance(one_or_more_links, list):
				for link in one_or_more_links:
					if isinstance(link, Link):
						yield wikipedia.get_page(url=link.url)
			elif isinstance(one_or_more_links, Link):
				yield wikipedia.get_page(url=one_or_more_links.url)

	if name == 'countries':
		page = wikipedia.get_page(url='https://en.wikipedia.org/wiki/List_of_national_capitals')
		table = page['tables'][0]

		def row_to_pensieve(row):
			pensieve = Pensieve(safe=False, warn_unsafe=False)
			pensieve['capital'] = row['city'][0] if isinstance(row['city'], list) else row['city']
			pensieve['country'] = row['country'][0] if isinstance(row['country'], list) else row['country']
			pensieve.store(key='capital_name', precursors=['capital'], function=lambda x: x.text, evaluate=False)
			pensieve.store(
				key='capital_page', precursors=['capital'], function=lambda x: wikipedia.get_page(url=x.url), evaluate=False
			)
			pensieve.store(key='country_name', precursors=['country'], function=lambda x: x.text, evaluate=False)
			pensieve.store(
				key='country_page', precursors=['country'],
				function=lambda x: wikipedia.get_page(url=x.url), evaluate=False
			)
			return pensieve
		return list(ProgressBar.apply(function=row_to_pensieve, data=table, echo=echo))





def find_common_left(strings):
	"""
	:param list[str] strings: list of strings we want to find a common left part in
	:rtype: str
	"""
	length = min([len(s) for s in strings])
	result = ''
	for i in range(length):
		if all([strings[0][i] == s[i] for s in strings[1:]]):
			result += strings[0][i]
		else:
			break
	return result


def find_common_right(strings):
	reversed_string = [s[::-1] for s in strings]
	return find_common_left(reversed_string)[::-1]


def find_common(strings, side='left'):
	"""
	:param list[str] strings: list of strings we want to find a common left part in
	:type side: str
	:rtype: str
	"""
	if side[0].lower() == 'l':
		return find_common_left(strings=strings)
	else:
		return find_common_right(strings=strings)



from .BasicMask import BasicMask


class Mask(BasicMask):
	def __add__(self, other):
		return self.combine(others=other)

	def combine(self, others):
		"""
		:param BasicMask or list[BasicMask] others: another mask or other masks
		:rtype: MaskSet
		"""
		if isinstance(others, self.__class__):
			others = [others]
		return MaskSet(masks=[self] + others)


class MaskSet:

	def __init__(self, masks=None):
		"""
		:param list[tuple] or list[BasicMask] masks: prefixes and suffixes in tuples
		"""

		self._masks = []

		masks = masks or []
		for x in masks:
			self.append(x=x)

	def __getstate__(self):
		return self._masks

	def __setstate__(self, state):
		self._masks = state

	def __repr__(self):
		return '\n'.join([str(x) for x in self._masks])

	def __str__(self):
		return self.__repr__()

	def train(self, string, target, inplace=True):
		print('training')
		"""
		:param str string: a string that might contain a target
		:param str target: the substring that might be contained in the string
		:param bool inplace: if True this maskset will be trained, otherwise a new maskset will be returned
		:rtype: MaskSet or NoneType
		"""
		if target not in string:
			raise ValueError('the string does not contain the target!')

		mask_set = self.__class__()
		splits = str(string).split(str(target))
		for i in range(len(splits)-1):
			mask_set.append((splits[i], splits[i+1]))
		print('mask_set:', mask_set, '\n')
		if inplace:
			if len(self._masks) == 0:
				self._masks = mask_set._masks
			else:
				self.intersect(others=mask_set, inplace=True)
		else:
			if len(self._masks) == 0:
				return mask_set
			else:
				return self.intersect(others=mask_set, inplace=False)

	def copy(self):
		return self.__class__(masks=self._masks)

	def append(self, x):
		if isinstance(x, BasicMask):
			mask = x.copy()
		else:
			mask = BasicMask(prefix=x[0], suffix=x[1])
		if mask not in self._masks and mask._prefix != '' and mask._suffix != '':
			self._masks.append(mask)

	def intersect(self, others, inplace=False):
		"""
		:type others: MaskSet or list[MaskSet]
		:param bool inplace: if True, this object will be affected, otherwide a new one is returned
		:rtype: MaskSet
		"""

		if isinstance(others, MaskSet):
			others = [others]

		result = self if inplace else self.copy()

		for other in others:
			intersection = self.__class__(masks=[])
			for mask in result._masks:
				for other_mask in other._masks:
					intersection.append(mask * other_mask)
			result._masks = intersection._masks

		if not inplace:
			return result

	def __mul__(self, other):
		"""
		:param list[MaskSet] or MaskSet other: other mask
		:rtype: MaskSet
		"""
		return self.intersect(others=other, inplace=False)



from bs4 import Tag
import re


def get_wikipedia_id(url_response):
	"""
	:type element: Tag
	:rtype: int
	"""
	keyword = '"wgArticleId"'
	text = url_response
	if keyword in text:
		beginning = text[text.find(keyword) + len(keyword):]
		ending = beginning[:beginning.find(',')]
		ints = re.findall('\d+', ending)
		if len(ints) > 0:
			return int(ints[0])

def get_page_title(url_response):
	"""
	:type element: Tag
	:rtype: int
	"""
	keyword = '"wgTitle":'
	text = url_response.replace('\\u0026', 'and')
	if keyword in text:
		beginning = text[text.find(keyword) + len(keyword):]
		ending = beginning[:beginning.find('",')]
		results = re.findall('".+', ending)
		if len(results) > 0:
			return results[0][1:]
		else: # try regex:
			results = re.findall('"wgTitle"\s*?:\s*?".*?"', text)
			if len(results) > 0:
				return results[0][11:-1]
			else:
				return 'No Title'
	else:
		return 'No Title'

def get_page_language(url_response):
	"""
	:type element: Tag
	:rtype: int
	"""
	keyword = '"wgPageContentLanguage"'
	text = url_response
	if keyword in text:
		beginning = text[text.find(keyword) + len(keyword):]
		ending = beginning[:beginning.find(',')]
		results = re.findall('".+"', ending)
		if len(results) > 0:
			return results[0][1:-1]

def get_page_namespace(url_response):
	"""
	:type element: Tag
	:rtype: int
	"""
	keyword = '"wgNamespaceNumber"'
	text = url_response
	if keyword in text:
		beginning = text[text.find(keyword) + len(keyword):]
		ending = beginning[:beginning.find(',')]
		ints = re.findall('\d+', ending)
		if len(ints) > 0:
			return int(ints[0])



import re


def is_wikipedia_page_url(url):
	wikipedia_url_regex_str = '^(http|https)://.+\.wikipedia.org/'
	wikipedia_url_regex = re.compile(wikipedia_url_regex_str)
	if re.match(wikipedia_url_regex, url):
		return True
	else:
		return False


def is_mobile_wikipedia_page_url(url):
	wikipedia_url_regex_str = '^(http|https)://.+\./m\.wikipedia.org/'
	wikipedia_url_regex = re.compile(wikipedia_url_regex_str)
	if re.match(wikipedia_url_regex, url):
		return True
	else:
		return False


def convert_mobile_wikipedia_page_url_to_normal_page(url):
	return url.replace('.m.wikipedia.org', '.wikipedia.org')



from .SearchResults import BingSearchResults, YahooSearchResults

from chronometry import MeasurementSet, get_elapsed, get_now

import requests
from requests.utils import quote
import warnings
import time


class SearchEngine:
	def __init__(self, rate_limit_wait_seconds=0.01, cache=None):
		"""
		:param float rate_limit_wait_seconds: wait between requests
		:param disk.cache_class.Cache cache:
		"""
		self._rate_limit_wait = rate_limit_wait_seconds
		self._rate_limit_last_call = None

		self._cache = cache

		if self._cache:
			self.request = self._cache.make_cached(
				id='search_engine_request_function',
				function=self._request,
				condition_function=self._request_result_valid,
				sub_directory='request'
			)

		else:
			self.request = self._request

		self._function_durations = MeasurementSet()

	@staticmethod
	def get_bing_search_url(query, site=None):
		"""
		:type query: str
		:type site: str
		:rtype: str
		"""
		full_query = quote(query).replace('%20', '+')
		if site is not None:
			full_query += f'+site:{site}'
		return 'https://www.bing.com/search?q=' + full_query

	def search_bing(self, query, site=None):
		return BingSearchResults(api=self, query=query, site=site)

	@staticmethod
	def get_yahoo_search_url(query, num_results=None, site=None):
		"""
		:type query: str
		:type site: str
		:rtype: str
		"""
		full_query = ''
		if num_results is not None:
			full_query += f'n={num_results}&'

		full_query += 'p='+quote(query).replace('%20', '+')

		if site is not None:
			query += f'&vs:{site}'

		return  'https://search.yahoo.com/search?' + full_query

	def search_yahoo(self, query, site=None, num_results=None):
		return YahooSearchResults(api=self, query=query, num_results=num_results, site=site)

	@staticmethod
	def _request_result_valid(response):
		if response.status_code == 200:
			return True
		else:
			warnings.warn(f'response status code: {response.status_code}')
			return False

	def _request(self, url=None, header=None):
		"""
		:type parameters: dict
		:rtype: dict
		"""
		if url is None:
			raise ValueError('url cannot be empty for non-json request!')

		# headers = {'User-Agent': self._user_agent}

		if self._rate_limit_wait and self._rate_limit_last_call:
			wait_time = self._rate_limit_wait - get_elapsed(start=self._rate_limit_last_call, unit='s')
			if wait_time > 0:
				time.sleep(wait_time)
		with warnings.catch_warnings():
			warnings.filterwarnings('ignore')
			result = requests.get(url, verify=False)

		if self._rate_limit_wait:
			self._rate_limit_last_call = get_now()
		return result



import requests
from requests.adapters import SSLError
from time import sleep
import warnings
from bs4 import BeautifulSoup
from disk import Cache, Path
from chronometry import get_now, get_elapsed


class Scraper:
	def __init__(self, name, expire_in=None, rate_limit_wait_seconds=0.01, cache=None, num_request_tries=4):
		if isinstance(cache, (str, Path)):
			cache = Cache(path=cache)

		self._name = name
		self._rate_limit_wait = rate_limit_wait_seconds
		self._rate_limit_last_call = None
		self._num_request_tries = num_request_tries
		self._expire_in = expire_in
		self._cache = cache
		self._create_cached_functions()

	def _create_cached_functions(self):
		if self.cache:
			self.request = self.cache.make_cached(
				id=f'{self._name}_request',
				function=self._request,
				sub_directory=f'{self._name}_request',
				expire_in=self._expire_in
			)
			self.get_request_soup = self._get_request_soup
		else:
			self.request = self._request
			self.get_request_soup = self._get_request_soup

	def _get_state_attribute_names(self):
		return ['_name', '_rate_limit_wait', '_rate_limit_last_call', '_num_request_tries', '_expire_in', '_cache']

	def __getstate__(self):
		return {attribute_name: getattr(self, attribute_name) for attribute_name in self._get_state_attribute_names()}

	def __setstate__(self, state):
		for key, value in state.items():
			setattr(self, key, value)
		self._create_cached_functions()

	@property
	def cache(self):
		"""
		:rtype: Cache or NoneType
		"""
		return self._cache

	def _request(self, url, verify=False):
		error = None
		for i in range(1, self._num_request_tries + 1):
			try:
				if self._rate_limit_wait and self._rate_limit_last_call:
					wait_time = self._rate_limit_wait - get_elapsed(start=self._rate_limit_last_call, unit='s')
					if wait_time > 0:
						sleep(wait_time)
				with warnings.catch_warnings():
					warnings.simplefilter('ignore')
					result = requests.get(url, verify=verify)
				break
			except SSLError as error:
				print(f'try {i}, error with get request with url="{url}"')
				warnings.warn(str(e))
				sleep(0.001 * 10 ** i)
		else:
			print(f'failed after {self._num_request_tries} requests!')
			raise error

		if self._rate_limit_wait:
			self._rate_limit_last_call = get_now()

		return result

	def _get_request_soup(self, content, features='lxml'):
		return BeautifulSoup(content, features)

	def get_soup(self, url, verify=False):
		"""
		:type url: str
		:rtype: BeautifulSoup
		"""
		return self.get_request_soup(content=self.request(url=url, verify=verify).content)



from imdb import IMDb
from imdb.Movie import Movie

_IMDb = IMDb()


class IMDB:
	def __init__(self, cache=None):
		self._imdb = _IMDb
		self._cache = cache
		if self._cache:
			self.search = self._cache.make_cached(
				id='imdb_search',
				function=self._search,
				condition_function=self._search_result_valid,
				sub_directory='search'
			)
		else:
			self.search = self._search

	def _search(self, title, year=None):
		"""
		:type title: str
		:type year: int
		:rtype: Movie
		"""
		movies = self._imdb.search_movie(title=title)

		if len(movies) == 0:
			print(f'{title} ({year}) not found!')
			return False

		if year is not None:
			for movie in movies:
				if movie.data['year'] == year:
					return self._imdb.get_movie(movie.movieID)

		return self._imdb.get_movie(movies[0].movieID)

	@staticmethod
	def _search_result_valid(result):
		if result is False:
			return False
		else:
			return True



# p
from copy import deepcopy
import requests
from requests.adapters import SSLError
import time
import warnings
import re

# i
from chronometry import MeasurementSet, get_elapsed, get_now
from abstract import Graph

from .exceptions import HTTPTimeoutError, WikipediaException
from .WikipediaPage import WikipediaPage
from .get_special_data import get_special_data


class Wikipedia:
	def __init__(
			self, language='en',
			user_agent='wikipedia (https://github.com/goldsmith/Wikipedia/)',
			rate_limit_wait_seconds=0.01,
			cache=None,
			num_request_tries=4
	):
		"""
		:param str language: such as 'en'
		:param str user_agent:
		:param float rate_limit_wait_seconds: wait between requests
		:param disk.Cache.Cache cache:
		"""
		self._language = language
		self._user_agent = user_agent
		self._rate_limit_wait = rate_limit_wait_seconds
		self._rate_limit_last_call = None
		self._num_request_tries = num_request_tries

		self._cache = cache
		if self._cache:
			self.request = self._cache.make_cached(
				id='wikipedia_request_function',
				function=self._request,
				condition_function=self._request_result_valid,
				sub_directory='request'
			)

		else:
			self.request = self._request

		self._function_durations = MeasurementSet()

	def __hashkey__(self):
		return (self.__class__.__name__, self._language, self._user_agent, self._rate_limit_wait, self._cache)

	def __getstate__(self):
		return {
			'language': self._language,
			'user_agent': self._user_agent,
			'rate_limit_wait': self._rate_limit_wait,
			'rate_limit_last_call': self._rate_limit_last_call,
			'cache': self._cache,
			'function_durations': self._function_durations
		}

	def __setstate__(self, state):
		self._language = state['language']
		self._user_agent = state['user_agent']
		self._rate_limit_wait = state['rate_limit_wait']
		self._rate_limit_last_call = state['rate_limit_last_call']
		self._cache = state['cache']
		self._function_durations = state['function_durations']
		if self._cache:
			self.request = self._cache.make_cached(
				id='wikipedia_request_function',
				function=self._request,
				condition_function=self._request_result_valid,
				sub_directory='request'
			)

		else:
			self.request = self._request

	def __eq__(self, other):
		"""
		:type other: Wikipedia
		:rtype: bool
		"""
		if isinstance(other, self.__class__):
			return self._language == other._language
		else:
			return False

	@property
	def cache(self):
		"""
		:rtype: disk.Cache or NoneType
		"""
		return self._cache

	@property
	def function_durations(self):
		"""
		:rtype: MeasurementSet
		"""
		return self._function_durations

	@property
	def language(self):
		return self._language.lower()

	@property
	def api_url(self):
		return 'http://' + self.language + '.wikipedia.org/w/api.php'

	@staticmethod
	def _request_result_valid(result, **kwargs):
		return result is not None

	def _request(self, parameters=None, url=None, format='json'):
		"""
		:type parameters: dict
		:rtype: dict
		"""
		_format = format
		for i in range(1, self._num_request_tries + 1):

			try:
				if _format == 'json':
					if parameters is None:
						raise ValueError('parameters cannot be empty for json request!')
					parameters['format'] = 'json'
					if 'action' not in parameters:
						parameters['action'] = 'query'
				else:
					if url is None:
						raise ValueError('url cannot be empty for non-json request!')

				headers = {'User-Agent': self._user_agent}

				if self._rate_limit_wait and self._rate_limit_last_call:
					wait_time = self._rate_limit_wait - get_elapsed(start=self._rate_limit_last_call, unit='s')
					if wait_time > 0:
						time.sleep(wait_time)

				if _format == 'json':
					r = requests.get(self.api_url, params=parameters, headers=headers)
					result = r.json()

				else:
					result = requests.get(url, headers=headers)

				break

			except SSLError as e:
				print(f'try {i}, error with get request with url="{url}", headers="{headers}", format="{_format}"')
				warnings.warn(str(e))
				time.sleep(0.001 * 10 ** i)


		else:
			raise e

			# result = html.document_fromstring(r.text)
			# result = r.text

		if self._rate_limit_wait:
			self._rate_limit_last_call = get_now()
		return result

	def get_page(self, url=None, id=None, title=None, namespace=0, redirect=True):
		"""
		:type id: int or str or NoneType
		:type url: str or NoneType
		:type title: str or NoneType
		:rtype: WikipediaPage
		"""
		return WikipediaPage(id=id, url=url, title=title, namespace=namespace, wikipedia=self, redirect=redirect)

	def get_page_graph(
			self, graph=None, id=None, url=None, title=None, namespace=0, redirect=True,
			max_depth=1, strict=True, ordering=True, echo=1
	):
		try:
			if graph:
				graph = deepcopy(graph)
			else:
				graph = Graph(obj=None, strict=strict, ordering=ordering)

			def _crawl(_graph, _page, _parent_page, _max_depth, _depth, _echo, _crawl_completed):
				if _page['url'] not in _graph:
					_graph.add_node(name=_page['url'], label=_page['title'], value=_page)

				if _parent_page is not None:
					_graph.connect(start=_parent_page['url'], end=_page['url'], if_edge_exists='ignore')

				# to avoid crawling the children of a page for a second time we add the url of the parent page in
				# crawl_completed at the end
				if _page['url'] not in _crawl_completed and _depth < _max_depth:
					for child in _page.get_children(echo=_echo):
						_crawl(
							_graph=_graph, _page=child,
							_parent_page=_page, _max_depth=_max_depth, _depth=_depth + 1, _echo=_echo,
							_crawl_completed=_crawl_completed
						)
					_crawl_completed.append(_page['url'])

			page = self.get_page(id=id, url=url, title=title, namespace=namespace, redirect=redirect)
			_crawl(
				_graph=graph, _page=page, _parent_page=None,
				_max_depth=max_depth, _echo=echo, _depth=0, _crawl_completed=[]
			)
			return graph
		except KeyboardInterrupt:
			warnings.warn('get_page_graph was interrupted by keyboard!')
			return graph

	def search(self, query, num_results=10, redirect=True):
		"""
		Do a Wikipedia search for `query`.
		:type query: str
		:param int num_results: the maxmimum number of results returned
		:type redirect: bool
		"""

		search_params = {
			'list': 'search',
			'srprop': '',
			'srlimit': num_results,
			'limit': num_results,
			'srsearch': query
		}

		raw_results = self.request(search_params)

		if 'error' in raw_results:
			if raw_results['error']['info'] in ('HTTP request timed out.', 'Pool queue is full'):
				raise HTTPTimeoutError(query)
			else:
				raise WikipediaException(raw_results['error']['info'])

		results = raw_results['query']['search']
		try:
			pages = [
				WikipediaPage(wikipedia=self, id=d['pageid'], title=d['title'], namespace=d['ns'], redirect=redirect) for d in results
			]
		except Exception as e:
			print('\n'*5, 'error in:\n', results, '\n'*5)
			raise e

		already_captured_urls = [page['url'] for page in pages]
		disambiguation_pages = [page for page in pages if page['disambiguation']]
		disambiguation_results = []
		total_num_results = len(pages)
		for disambiguation_page in disambiguation_pages:
			for link in disambiguation_page['disambiguation_results']:
				if link.url not in already_captured_urls and total_num_results<num_results:
					wikipedia_url_regex_str = '^(http|https)://.+\.wikipedia.org'
					wikipedia_url_regex = re.compile(wikipedia_url_regex_str)
					if re.match(wikipedia_url_regex, link['url']):
						page = WikipediaPage(
							wikipedia=self, url=link.url, title=link.text,
							disambiguation_url=disambiguation_page['url']
						)
						disambiguation_results.append(page)
						total_num_results += 1
		return pages + disambiguation_results

	def get_performance(self):
		return self.function_durations.summary_data

	def get_data(self, name, echo=1):
		return get_special_data(wikipedia=self, name=name, echo=echo)

WIKIPEDIA = Wikipedia()



import requests
from disk import Cache
from requests.adapters import SSLError
from time import sleep
from chronometry import get_now, get_elapsed
import warnings


class Web:
	def __init__(
			self, id=0, cache=None, expire_in=None, num_request_tries=4, rate_limit_wait_seconds=0.001,
			headers=None, parameters=None
	):
		if cache is None:
			self._cache = Cache(path='internet_cache')
		elif isinstance(cache, str):
			self._cache = Cache(path=cache)
		elif isinstance(cache, Cache):
			self._cache = cache
		elif cache == False:
			self._cache = None

		self._expire_in = expire_in
		self._id = id
		self._num_request_tries = num_request_tries
		self._rate_limit_wait = rate_limit_wait_seconds
		self._headers = headers
		self._parameters = parameters

	def _create_cached_functions(self):
		if self.cache:
			self.request = self.cache.make_cached(
				id=f'{self._id}_request',
				function=self._request,
				sub_directory=f'{self._id}_request',
				expire_in=self._expire_in
			)
		else:
			self.request = self._request

	@property
	def cache(self):
		"""
		:rtype: NoneType or Cache
		"""
		return self._cache

	def _request(self, url, verify=False, headers=None, parameters=None):
		headers = headers or self._headers
		parameters = parameters or self._parameters
		error = None
		for i in range(self._num_request_tries):
			if i > 0:
				sleep(0.001 * 10 ** (i-1))
			try:
				if self._rate_limit_wait and self._rate_limit_last_call:
					wait_time = self._rate_limit_wait - get_elapsed(start=self._rate_limit_last_call, unit='s')
					if wait_time > 0:
						sleep(wait_time)
				with warnings.catch_warnings():
					warnings.simplefilter('ignore')
					result = requests.get(url, params=parameters, headers=headers, verify=verify)
				break
			except SSLError as caught_error:
				print(f'try {i + 1}, error with get request with url="{url}"')
				warnings.warn(str(caught_error))
				error = caught_error
		else:
			print(f'failed after {self._num_request_tries} requests!')
			if error is None:
				raise RuntimeError(f'failed after {self._num_request_tries} requests!')
			else:
				raise error

		if self._rate_limit_wait:
			self._rate_limit_last_call = get_now()

		return result






from .exceptions import PageError, RedirectError, ODD_ERROR_MESSAGE
from .InfoBox import InfoBox
from .get_wikipedia_id import get_wikipedia_id, get_page_title, get_page_language, get_page_namespace
from .Page_helpers import *
from .is_wikipedia_page_url import is_wikipedia_page_url
from .is_wikipedia_page_url import is_mobile_wikipedia_page_url
from .is_wikipedia_page_url import convert_mobile_wikipedia_page_url_to_normal_page

import re
from pensieve import Pensieve
from slytherin.collections import remove_list_duplicates, flatten
from ravenclaw.wrangling import standardize_columns
from chronometry.progress import ProgressBar

from silverware import Spoon, Link
from bs4 import BeautifulSoup
from pandas import DataFrame
import warnings


class WikipediaPage:
	def __init__(
			self, wikipedia, id=None, url=None, title=None, namespace=None, redirect=True, disambiguation_url=None,
			ignore_error=False
	):
		self._wikipedia = wikipedia
		self._id = id
		self._url = url
		self._title = title
		self._ignore_error = ignore_error
		self._setup_pensieve()
		self.pensieve['namespace'] = namespace
		self.pensieve['redirect'] = redirect
		self.pensieve['disambiguation_url'] = disambiguation_url
		self._load_primary()

		try:
			self._load_the_rest()
		except Exception as e:
			warnings.warn(f'failed to load the rest id: "{self._id}", url: "{self._url}", title: "{self._title}"')
			if not ignore_error:
				raise e

	_STATE_ATTRIBUTES_ = ['_id', '_url', '_title', '_ignore_error']

	@property
	def pensieve(self):
		"""
		:rtype: Pensieve
		"""
		return self._pensieve

	def __getstate__(self):
		state = {key: getattr(self, key) for key in self._STATE_ATTRIBUTES_}
		state['_pensieve'] = self._pensieve.get_contents()
		return state

	def __setstate__(self, state):
		for key, value in state.items():
			setattr(self, key, value)
		self._pensieve = state['_pensieve']
		self._load_primary()
		try:
			self._load_the_rest()
		except Exception as e:
			warnings.warn(f'failed to load the rest id: "{self._id}", url: "{self._url}", title: "{self._title}"')
			if not self._ignore_error:
				raise e

	def _setup_pensieve(self):
		self._pensieve = Pensieve(
			safe=False, function_durations=self.wikipedia.function_durations, warn_unsafe=False, hide_ignored=False
		)

	def _load_primary(self):
		if self._id or self._title or self._url:
			pass
		else:
			raise ValueError('Either id or title or url should be given!')
		if self._id:
			self.pensieve['original_id'] = self._id
			try:
				self._load_from_id()
			except Exception as e:
				warnings.warn(f'failed to load from id: "{self._id}"')
				if not self._ignore_error:
					raise e

		elif self._url:
			if is_wikipedia_page_url(url=self._url):
				if is_mobile_wikipedia_page_url(url=self._url):
					url = convert_mobile_wikipedia_page_url_to_normal_page(url=self._url)
				else:
					url = self._url

				self.pensieve['url'] = url
				try:
					self._load_from_url()
				except Exception as e:
					warnings.warn(f'failed to load from url: "{self._url}"')
					if not self._ignore_error:
						raise e
			else:
				raise ValueError(f'{self._url} does not match the wikipedia page pattern!')

		elif self._title:
			self.pensieve['original_title'] = self._title
			try:
				self._load_from_title()
			except Exception as e:
				warnings.warn(f'failed to load from title: "{self._title}"')
				if not self._ignore_error:
					raise e

	@property
	def wikipedia(self):
		"""
		:rtype: .Wikipedia_class.Wikipedia
		"""
		if self._wikipedia is None:
			raise AttributeError('Wikipedia API is missing!')
		else:
			return self._wikipedia

	def __eq__(self, other):
		"""
		:type other: WikipediaPage
		:rtype: bool
		"""
		if isinstance(other, self.__class__):
			return self['url'] == other['url']
		else:
			return False

	def __str__(self):
		if 'url' in self.pensieve:
			url = self.pensieve['url']
			return f'{self.title}: {url} '
		else:
			return f'{self.title}: {self.id} '

	def __repr__(self):
		return str(self)

	def __getitem__(self, item):
		return self.pensieve[item]

	def __graph__(self):
		return self.pensieve.__graph__()

	@property
	def title(self):
		"""
		:rtype: str
		"""
		if 'title' in self.pensieve:
			return self.pensieve['title']
		else:
			return self.pensieve['original_title']

	@property
	def id(self):
		"""
		:rtype: int
		"""
		if 'id' in self.pensieve:
			return self.pensieve['id']
		else:
			return self.pensieve['original_id']

	@property
	def url(self):
		if 'url' not in self.pensieve:
			raise AttributeError(f'Page {self} does not have a url!')
		elif self.pensieve['url'] is None:
			raise AttributeError(f'Page {self} does not have a url!')
		return self.pensieve['url']

	@property
	def base_url(self):
		"""
		:rtype: str
		"""
		return 'http://' + self.wikipedia.language + '.wikipedia.org'

	@wikipedia.setter
	def wikipedia(self, wikipedia):
		"""
		:type wikipedia: .Wikipedia_class.Wikipedia
		"""
		self._wikipedia = wikipedia

	def get_children(self, echo=1):
		link_lists = self['link_list']
		if link_lists:
			urls = remove_list_duplicates([link.url for link in flatten(link_lists)])
			wikipedia_urls = [url for url in urls if re.match('^https://.+\.wikipedia.org/', url)]
			non_php_urls = [url for url in wikipedia_urls if '/index.php?' not in url]

			pages = ProgressBar.map(
				function=lambda x: self.__class__(url=x, redirect=self['redirect'], wikipedia=self.wikipedia),
				iterable=non_php_urls, echo=echo, text=self['url']
			)
			return pages
		else:
			return []

	def request(self, url=None, parameters=None, format='html'):
		return self.wikipedia.request(url=url, parameters=parameters, format=format)

	def clear(self):
		new_pensieve = Pensieve(safe=True)
		for key in ['original_id', 'original_title', 'namespace', 'redirect', 'redirected_from']:
			new_pensieve[key] = self.pensieve[key]
		self._pensieve = new_pensieve

	def _search_page(self, id, title, redirect, redirected_from, num_recursions=0):
		if num_recursions > 3:
			raise RecursionError()
		# print(dict(title=title, id=id, redirect=redirect, redirected_from=redirected_from, num_recursions=num_recursions))

		search_query_parameters = get_search_parameters(id=id, title=title)
		search_request = self.request(parameters=search_query_parameters, format='json')

		query = search_request['query']

		id = list(query['pages'].keys())[0]
		page = query['pages'][id]
		title = page['title']
		full_url = page['fullurl']
		language = page['pagelanguage']
		namespace = page['ns']

		# missing is present if the page is missing

		if 'missing' in page:
			raise PageError(id=id, title=title)

		# same thing for redirect, except it shows up in query instead of page for
		# whatever silly reason
		elif 'redirects' in query:
			if redirect:
				redirects = query['redirects'][0]

				if 'normalized' in query:
					normalized = query['normalized'][0]
					assert normalized['from'] == self.title, ODD_ERROR_MESSAGE

					from_title = normalized['to']

				else:
					from_title = self.title

				assert redirects['from'] == from_title, ODD_ERROR_MESSAGE

				# change the title and reload the whole object

				return self._search_page(
					id=id, title=redirects['to'],
					redirect=redirect, redirected_from=redirects['from'],
					num_recursions=num_recursions+1
				)
			else:
				raise RedirectError(getattr(self, 'title', page['title']))

		# since we only asked for disambiguation in pageprop,
		# if a pageprop is returned,
		# then the page must be a disambiguation page
		elif 'pageprops' in page:
			return {
				'id': int(id), 'title': title, 'page': page, 'redirected_from': redirected_from,
				'full_url': full_url, 'language': language, 'namespace': namespace, 'disambiguation': True
			}

		else:
			return {
				'id': int(id), 'title': title, 'page': page, 'redirected_from': redirected_from,
				'full_url': full_url, 'language': language, 'namespace': namespace, 'disambiguation': False
			}

	def _get_url_response(self):
		self.pensieve.store(
			key='url_response', precursors='url', evaluate=False,
			function=lambda x: self.request(url=x, format='response')
		)

	def _load_from_url(self):
		self._get_url_response()

		self.pensieve.store(
			key='original_id', precursors=['url_response'], evaluate=False,
			function=lambda x: get_wikipedia_id(x.text)
		)

		self.pensieve.store(
			key='search_result', precursors=['original_id', 'redirect'], evaluate=False,
			function=lambda x: self._search_page(
				title=None, id=x['original_id'], redirect=x['redirect'],
				redirected_from=None
			)
		)

		self.pensieve.store(
			key='id', precursors=['url_response'], evaluate=False,
			function=lambda x: get_wikipedia_id(x.text)
		)
		self.pensieve.store(
			key='title', precursors=['url_response'], evaluate=False,
			function=lambda x: get_page_title(x.text)
		)
		self.pensieve.store(
			key='language', precursors=['url_response'], evaluate=False,
			function=lambda x: get_page_language(x.text)
		)
		self.pensieve.store(
			key='namespace', precursors=['url_response'], evaluate=False,
			function=lambda x: get_page_namespace(x.text)
		)
		self.pensieve.store(
			key='full_url', precursors=['url'], evaluate=False,
			function=lambda x: x
		)
		self.pensieve['disambiguation'] = False
		self.pensieve['redirected_from'] = None

	def _load_from_id(self):
		self.pensieve.store(
			key='search_result', precursors=['original_id', 'redirect'], evaluate=False,
			function=lambda x: self._search_page(
				title=None, id=x['original_id'], redirect=x['redirect'],
				redirected_from=None
			)
		)
		self.pensieve.decouple(key='search_result', prefix='')

		try:
			self.pensieve.store(
				key='json', precursors=['id', 'title'], evaluate=False,
				function=lambda x: self._get_json(id=x['id'], title=x['title'])
			)
		except Exception as e:
			display(self.pensieve)
			raise e

		self.pensieve.store(key='url', precursors=['page'], function=lambda x: x['fullurl'], evaluate=False)
		self._get_url_response()

	def _load_from_title(self):
		self.pensieve.store(
			key='search_result', precursors=['original_title', 'redirect'], evaluate=False,
			function=lambda x: self._search_page(
				title=x['original_title'], id=None, redirect=x['redirect'],
				redirected_from=None
			)
		)
		self.pensieve.decouple(key='search_result')
		self.pensieve.store(
			key='json', precursors=['id', 'title'], evaluate=False,
			function=lambda x: self._get_json(id=x['id'], title=x['title'])
		)
		self.pensieve.store(
			key='url', precursors=['page'],
			function=lambda x: x['fullurl'], evaluate=False
		)
		self._get_url_response()

	def _load_the_rest(self):
		self.pensieve.store(
			key='base_url', precursors=['url'],
			evaluate=False, lazy=True,
			function=lambda x: x[:x.find('/wiki/')]
		)

		# main parts
		def _add_see_also_flag(x):
			x = re.sub(
				r'<h2>.*>(references|external\s+links|notes\s+and\s+references|see\s+also)<.*</h2>',
				'<h2>SEEALSO</h2><ul><li><a href="http://SEEALSO" title="SEEALSO">SEEALSO</a></li></ul>',
				x,
				flags=re.IGNORECASE
			)
			return x

		def _get_beautiful_soup(url_response):
			text = _add_see_also_flag(url_response.text.replace('\\u0026', 'and').replace('&amp;', 'and'))
			soup = BeautifulSoup(text, 'lxml')
			for element in soup.find_all(name='div', attrs={'id': 'mw-panel'}):
				element.decompose()
			for element in soup.find_all(name='div', attrs={'id': 'mw-head'}):
				element.decompose()
			return soup


		self.pensieve.store(
			key='separated_body', precursors=['url_response'], evaluate=False,
			function=lambda url_response: separate_body_from_navigation_and_info_box(
				url_response=url_response
			)
		)

		self.pensieve.store(
			key='body',
			function=lambda separated_body: separated_body['body']
		)

		self.pensieve.store(
			key='headers', precursors=['body'], evaluate=False,
			function=lambda x: x.find_all(['h1', 'h2', 'h3'])
		)

		self.pensieve.store(
			key='info_box', evaluate=False,
			function=lambda separated_body: InfoBox(separated_body['info_box'])
		)

		self.pensieve.store(
			key='vertical_navigation_box',  evaluate=False,
			function=lambda separated_body: separated_body['vertical_navigation_box']
		)

		self.pensieve.store(
			key='navigation_boxes', evaluate=False,
			function=lambda separated_body: [
				box for box in separated_body['navigation_boxes']
			]
		)

		self.pensieve.store(
			key='category_box', evaluate=False,
			function=lambda separated_body: separated_body['category_box']
		)

		# end of main parts

		self.pensieve.store(
			key='paragraphs', precursors=['body'], evaluate=False,
			function=lambda x: get_main_paragraphs(body=x)
		)

		self.pensieve.store(
			key='paragraph_links', precursors=['paragraphs', 'base_url'], evaluate=False,
			function=lambda x: [
				[
					link.url
					for link in Spoon.find_links(element=paragraph, base_url=x['base_url'])
					if isinstance(link, Link)
				]
				for paragraph in x['paragraphs']
			]
		)

		self.pensieve.store(
			key='category_links', evaluate=False,
			function=lambda category_box, base_url: Spoon.find_links(
				element=category_box, base_url=base_url
			)
		)

		self.pensieve.store(
			key='categories', evaluate=False,
			function=get_categories
		)

		self.pensieve.store(
			key='disambiguation_results',
			precursors=['disambiguation', 'body', 'base_url'],
			evaluate=False,
			function=lambda x: get_disambiguation_results(
				disambiguation=x['disambiguation'], html=x['body'], base_url=x['base_url']
			)
		)

		self.pensieve.store(
			key='tables',
			precursors=['body', 'base_url'],
			evaluate=False,
			function=lambda x: [
				standardize_columns(data=table)
				for table in Spoon.filter(
					soup=x['body'], name='table', attributes={'class': 'wikitable'}
				).read_tables(base_url=x['base_url'], parse_links=True)
			]
		)

		self.pensieve.store(
			key='table_links',
			precursors=['tables'],
			evaluate=False,
			function=find_main_links_in_tables
		)

		self.pensieve.store(
			key='link_and_anchors', precursors=['body', 'base_url'], evaluate=False,
			function=lambda x: get_anchors_and_links(soup=x['body'], base_url=x['base_url'])
		)

		self.pensieve.store(
			key='link_and_anchor_list', precursors=['link_and_anchors'], evaluate=False,
			function=lambda x: x['list_link_and_anchors']
		)

		self.pensieve.store(
			key='nested_link_and_anchor_lists', precursors=['body', 'base_url'], evaluate=False,
			function=lambda x: Spoon.get_lists(element=x['body'], links_only=True, base_url=x['base_url'])
		)

		# 	LINKS IN A PAGE
		def _remove_anchors(link_lists):
			if isinstance(link_lists, list):
				result = [_remove_anchors(x) for x in link_lists if x is not None]
				return [x for x in result if x is not None]
			elif isinstance(link_lists, Link):
				if link_lists.url.startswith('#'):
					return None
				else:
					return link_lists
			else:
				return link_lists

		# 	LINKS IN A PAGE
		def _remove_nonanchors(link_lists):
			if isinstance(link_lists, list):
				return [_remove_nonanchors(x) for x in link_lists if x is not None]
			elif isinstance(link_lists, Link):
				if not link_lists.url.startswith('#'):
					return None
				else:
					return link_lists
			else:
				return link_lists

		self.pensieve.store(
			key='nested_anchor_lists', precursors=['nested_link_and_anchor_lists'], evaluate=False,
			function=_remove_nonanchors
		)

		self.pensieve.store(
			key='anchor_list', precursors=['link_and_anchor_list'], evaluate=False,
			function=_remove_nonanchors
		)



		self.pensieve.store(
			key='nested_link_lists', precursors=['nested_link_and_anchor_lists'], evaluate=False,
			function=_remove_anchors
		)

		self.pensieve.store(
			key='link_list', precursors=['link_and_anchor_list'], evaluate=False,
			function=_remove_anchors
		)

		self.pensieve.store(
			key='summary', precursors=['id', 'title'], evaluate=False,
			function=lambda x: get_page_summary(page=self, id=x['id'], title=x['title'])
		)

		self.pensieve.store(
			key='content', precursors=['id', 'title'], evaluate=False,
			function=lambda x: self._get_content(id=x['id'], title=x['title'])
		)

		self.pensieve.store(
			key='extract', precursors=['content'], evaluate=False, lazy=True,
			function=lambda x: x['extract']
		)

		self.pensieve.store(
			key='revision_id', precursors=['content'], evaluate=False, lazy=True,
			function=lambda x: x['revisions'][0]['revid']
		)

		self.pensieve.store(
			key='parent_id', precursors=['content'], evaluate=False, lazy=True,
			function=lambda x: x['revisions'][0]['parentid']
		)

	def keys(self):
		return self.pensieve.keys()

	def _get_json(self, id, title):
		id = str(id)
		html_query_parameters = get_html_parameters(id=id, title=title)
		html_request = self.request(parameters=html_query_parameters, format='json')
		return html_request['query']['pages'][id]['revisions'][0]['*']

	def _get_content(self, id, title):
		id = str(id)
		content_parameters = get_content_parameters(id=id, title=title)
		content_request = self.request(parameters=content_parameters, format='json')
		return content_request['query']['pages'][id]

	@property
	def body(self):
		"""
		:rtype: BeautifulSoup
		"""
		return self['body']

	@property
	def tables(self):
		"""
		:rtype: list[DataFrame]
		"""
		return self['tables']

	def __hashkey__(self):
		return (repr(self.url), repr(self.__class__))

	# def __hashkey__(self):
	#	return (self.__class__.__name__, tuple(getattr(self, name) for name in self._STATE_ATTRIBUTES_))


from silverware import find_links
from ..wikipedia import is_wikipedia_page_url
from .is_yahoo_finance_search_url_function import is_yahoo_finance_search_url

from bs4 import BeautifulSoup
from pensieve import Pensieve


class SearchResults:
	def __init__(self, api, query):
		self._api = api
		self._pensieve = Pensieve(safe=False, warn_unsafe=False)
		self._pensieve['query'] = query

	def _get_results(self):
		self._pensieve.store(
			key='response', precursors=['search_url'], function=lambda x: self.api.request(x),
			lazy=True, evaluate=False
		)
		self._pensieve.store(
			key='html', precursors=['response'], function=lambda x: x.text,
			evaluate=False
		)
		self._pensieve.store(
			key='parsed_html', precursors=['html'], function=lambda x: BeautifulSoup(x, 'lxml'),
			evaluate=False
		)

	@property
	def api(self):
		"""
		:rtype: .Search_class.Bing
		"""
		return self._api

	def __getitem__(self, item):
		return self._pensieve[item]


class BingSearchResults(SearchResults):
	def __init__(self, api, query, site=None):
		super().__init__(api=api, query=query)
		self._pensieve['site'] = site

		self._pensieve.store(
			key='search_url', precursors=['query', 'site'],
			function=lambda x: self.api.get_bing_search_url(query=x['query'], site=x['site'])
		)
		self._get_results()

		self._pensieve.store(
			key='links', precursors=['parsed_html'],
			function=lambda x: [
				link for link in find_links(element=x, base=None)
				if link.url.startswith('http://') or link.url.startswith('https://')
			],
			evaluate=False
		)

		self._pensieve.store(
			key='wikipedia_links', precursors=['links'],
			function=lambda x: [link for link in x if is_wikipedia_page_url(url=link['url'])]
		)


class YahooSearchResults(SearchResults):
	def __init__(self, api, query, num_results=None, site=None):
		super().__init__(api=api, query=query)

		self._pensieve['site'] = site
		self._pensieve['num_results'] = num_results

		self._pensieve.store(
			key='search_url', precursors=['query', 'num_results', 'site'],
			function=lambda x: self.api.get_yahoo_search_url(
				query=x['query'], num_results=x['num_results'], site=x['site']
			)
		)
		self._get_results()
		self._pensieve.store(
			key='links', precursors=['parsed_html'],
			function=lambda x: [
				link for link in find_links(element=x, base=None)
				if link.url.startswith('http://') or link.url.startswith('https://')
			],
			evaluate=False
		)

		self._pensieve.store(
			key='wikipedia_links', precursors=['links'],
			function=lambda x: [link for link in x if is_wikipedia_page_url(url=link.url)]
		)

		self._pensieve.store(
			key='yahoo_finance_search_urls', precursors=['links'],
			function=lambda x: [link.url for link in x if is_yahoo_finance_search_url(url=link.url)]
		)



import requests
import re
from requests_ntlm import HttpNtlmAuth
from getpass import getpass
import warnings


def get_id_token(url, username, password=None, num_tries=10):
    if password is None:
        credentials = HttpNtlmAuth(username=username, password=getpass())
    else:
        credentials = HttpNtlmAuth(username=username, password=password)
    session = requests.Session()
    token = None
    for i in range(num_tries):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            response = session.get(url, auth=credentials, verify=False, allow_redirects=False)
        url = response.headers['Location']
        try:
            token = re.search(r'id_token=(.+?)&', url).group(1)
            break
        except AttributeError:
            continue
    return token



from .find_common import find_common


class BasicMask:
	def __init__(self, prefix, suffix):
		"""
		:param str prefix: left side of the mask
		:param str suffix: right side of the mask
		"""
		self._prefix = str(prefix)
		self._suffix = str(suffix)

	def __getstate__(self):
		return self._prefix, self._suffix

	def __setstate__(self, state):
		self._prefix = state[0]
		self._suffix = state[1]

	def __eq__(self, other):
		"""
		:type other: BasicMask
		:rtype: bool
		"""
		return self._prefix == other._prefix and self._suffix == other._suffix

	def __ne__(self, other):
		return not self.__eq__(other)

	def __repr__(self):
		return f'"{self._prefix}", "{self._suffix}"'

	def __str__(self):
		return self.__repr__()

	def copy(self):
		return self.__class__(prefix=self._prefix, suffix=self._suffix)

	def is_in(self, string):
		"""
		:param str string: string to be check for existence of the mask
		:rtype: bool
		"""
		if self._prefix not in string:
			return False
		else:
			right_side = string.split(self._prefix, maxsplit=1)[1]
			return self._suffix in right_side

	def apply(self, string):
		"""
		:param str string:
		:rtype: list[str]
		"""
		if not self.is_in(string=string):
			return []

		else:
			split1 = string.split(sep=self._prefix)
			if string.startswith(self._prefix):
				to_the_right_of_prefix = split1
			else:
				to_the_right_of_prefix = split1[1:]

			sandwiched = [x for x in to_the_right_of_prefix if self._suffix in x]
			sandwiched = [x.split(self._suffix, maxsplit=1)[0] for x in sandwiched]
			return sandwiched

	def intersect(self, others):
		"""
		:param BasicMask or list[BasicMask] others: another mask or other masks
		:rtype: BasicMask
		"""
		if isinstance(others, self.__class__):
			others = [others]
		prefixes = [self._prefix] + [x._prefix for x in others]
		suffixes = [self._suffix] + [x._suffix for x in others]
		return self.__class__(
			prefix=find_common(strings=prefixes, side='right'),
			suffix=find_common(strings=suffixes, side='left')
		)

	def __mul__(self, other):
		return self.intersect(others=other)



from setuptools import setup, find_packages


def readme():
	with open('./README.md') as f:
		return f.read()


setup(
	name='cyberspace',
	version='2022.2.12',
	license='MIT',

	url='https://github.com/idin/cyberspace',
	author='Idin',
	author_email='py@idin.ca',

	description='Python library for navigating the world wide web',
	long_description=readme(),
	long_description_content_type='text/markdown',

	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Programming Language :: Python :: 3 :: Only',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Topic :: Software Development :: Libraries :: Python Modules'
	],

	packages=find_packages(exclude=["jupyter_tests", ".idea", ".git"]),
	install_requires=[
		'IMDbPy', 'bs4', 'requests_ntlm',
		'pandas', 'requests', 'memoria',
		'chronometry', 'slytherin', 'abstract', 'pensieve', 'ravenclaw', 'silverware'
	],
	python_requires='~=3.6',
	zip_safe=False
)



from silverware import Spoon, Link, parse_link, find
from bs4 import BeautifulSoup, Tag


def get_search_parameters(id, title):
	query_parameters = {
		'prop': 'info|pageprops',
		'inprop': 'url',
		'ppprop': 'disambiguation',
		'redirects': '',
	}

	if id:
		query_parameters['pageids'] = id
	else:
		query_parameters['titles'] = title

	return query_parameters


def get_disambiguation_results(disambiguation, html, base_url):
	if disambiguation:
		original_content = html.find(attrs={'id': 'bodyContent'}).find(attrs={'id': 'mw-content-text'})
		content = Spoon.clone(soup=original_content, in_spoon=False)
		disambigbox = content.find(attrs={'id': 'disambigbox'})
		if disambigbox:
			disambigbox.extract()
		printfooter = content.find(attrs={'class': 'printfooter'})
		if printfooter:
			printfooter.extract()
		links = Spoon.find_links(element=content, base_url=base_url)
		return [link for link in links if hasattr(link, 'url') and '/index.php?' not in link.url]
	else:
		return []


def get_html_parameters(id, title):
	query_params = {
		'prop': 'revisions',
		'rvprop': 'content',
		'rvparse': '',
		'rvlimit': 1
	}
	if id:
		query_params['pageids'] = id
	else:
		query_params['titles'] = title
	return query_params


def get_summary_parameters(id, title):
	query_params = {
		'prop': 'extracts',
		'explaintext': '',
		'exintro': '',
	}
	if id:
		query_params['pageids'] = id
	else:
		query_params['titles'] = title
	return query_params


def get_page_summary(page, id, title):
	id = str(id)
	summary_query_parameters = get_summary_parameters(id=id, title=title)
	summary_request = page.request(parameters=summary_query_parameters, format='json')
	try:
		return summary_request['query']['pages'][id]['extract']
	except Exception as e:
		print('error with', page, id, title, sep='\n')
		raise e

def get_content_parameters(id, title):
	query_params = {
		'prop': 'extracts|revisions',
		'explaintext': '',
		'rvprop': 'ids'
	}
	if id:
		query_params['pageids'] = id
	else:
		query_params['titles'] = title

	return query_params


def separate_body_from_navigation_and_info_box(url_response):
	html = BeautifulSoup(url_response.text, 'lxml')

	vertical_navigation_box = html.find(name='table', attrs={'class': 'vertical-navbox'})
	info_box = html.find(name='table', attrs={'class': 'infobox'})
	navigation_boxes = html.find_all(name='div', attrs={'role': 'navigation'})
	categories = html.find(name='div', attrs={'id': 'catlinks'})
	category_box = html.find(name='div', attrs={'id': 'catlinks'})
	external_links = html.find(name='span', attrs={'id': ['External_links', 'References', 'See_also', 'Notes_and_references']})
	if external_links:
		spoon = Spoon(soup=html).after(external_links)
		external_links = spoon.extract()

	result = {
		'body': html,
		'vertical_navigation_box': vertical_navigation_box,
		'info_box': info_box,
		'categories': categories,
		'navigation_boxes': navigation_boxes or [],
		'category_box': category_box,
		'external_links': external_links
	}

	if vertical_navigation_box:
		vertical_navigation_box.extract()
	if info_box:
		info_box.extract()
	for navigation_box in navigation_boxes:
		navigation_box.extract()
	if categories:
		categories.extract()
	if category_box:
		category_box.extract()

	return result


def get_categories(category_links):
	if isinstance(category_links, list):
		urls = [link.url for link in category_links if isinstance(link, Link)]
		category_sign = '/wiki/Category:'
		category_urls = [url for url in urls if category_sign in url]
		return [url[url.find(category_sign) + len(category_sign):] for url in category_urls]
	else:
		return []


def is_good_link(_link):
	unacceptable = [
		'/w/index.php?', '/wiki/Special:', '/wiki/Help:',
		'/wiki/Wikipedia:', 'https://foundation.wikimedia.org',
		'/wiki/Talk:', '/wiki/Portal:', '/shop.wikimedia.org',
		'www.mediawiki.org', '/wiki/Main_Page', 'wikimediafoundation.org',
		'/wiki/Template:', '/wiki/Template_talk:'
	]
	for x in unacceptable:
		if x in _link:
			return False
	return True


def find_links_in_tables(soup, base_url):
	result = {}
	for table in soup.find_all('table'):
		for item in table.find_all('li'):
			for link in item.find_all('a'):
				parsed_link = parse_link(element=link, base=base_url)
				if isinstance(parsed_link, Link):
					if parsed_link.url == 'http://SEEALSO':
						break
					result[id(link)] = parsed_link
			else:
				continue

			break

		else:
			continue

		break

	return result


def find_links_in_list(soup, base_url):
	result = {}
	for html_list in soup.find_all(['ol', 'ul']):
		for item in html_list.find_all('li'):
			children = item.children
			try:
				first_child = next(children)
			except:
				try:
					first_child = children[0]
				except:
					continue
			link = find(elements=first_child, name='a')
			if isinstance(first_child, (BeautifulSoup, Tag)):
				parsed_link = parse_link(element=link, base=base_url)
				if isinstance(parsed_link, Link):
					if parsed_link.url == 'http://SEEALSO':
						break
					result[id(link)] = parsed_link
		else:
			continue

		break

	return result


def get_anchors_and_links(soup, base_url):
	table_links = find_links_in_tables(soup=soup, base_url=base_url)
	links_in_lists = {
		i: link for i, link in find_links_in_list(soup=soup, base_url=base_url).items()
		if i not in table_links
	}
	return {
		'list_link_and_anchors': list(links_in_lists.values()),
		'table_links_and_anchors': list(table_links.values())
	}


def find_main_links_in_tables(tables):
	result = {}
	for table in tables:
		for col in table.columns:
			links = [x for x in table[col] if isinstance(x, Link)]
			link_lists = [
				maybe_link for lists in [x for x in table[col] if isinstance(x, list)]
				for maybe_link in lists if isinstance(maybe_link, Link)
			]
			links = links + link_lists
			if len(links) > 0:
				result[str(col)] = links

	return result


def get_main_paragraphs(body):
	_paragraphs_and_headers = body.find_all(['p', 'h1', 'h2', 'h3'])
	paragraphs = []
	for tag in _paragraphs_and_headers:
		if tag.name == 'p':
			paragraphs.append(tag)
		else:
			if tag.text == 'SEEALSO':
				break
	return paragraphs






from .find_common import find_common


class Template:
	def __init__(self):
		self._left = None
		self._right = None

	def train(self, targets, string):
		splits = [string.split(t) for t in targets]
		before_afters = {t:(s[:-1], s[1:]) for t in targets for s in string.split(t)}
		num_before_afters = len(splits[0])-1
		common_before_afters = [
			(
				find_common([ba[i][0] for ba in before_afters.values()], side='left'),
				find_common([ba[i][1] for ba in before_afters.values()], side='right')
			)
			for i in range(num_before_afters)
		]



